"""pypetwalk is a Python library to communicate with the petWALK.control module."""
from __future__ import annotations

import logging
from types import TracebackType

from .api import API
from .const import (
    API_METHOD_MAPPING,
    API_PORT,
    API_STATE_BRIGHTNESS_SENSOR,
    API_STATE_DOOR,
    API_STATE_MAPPING,
    API_STATE_MOTION_IN,
    API_STATE_MOTION_OUT,
    API_STATE_RFID,
    API_STATE_SYSTEM,
    API_STATE_TIME,
    WS_PORT,
)
from .exceptions import PyPetWALKInvalidResponse, PyPetWALKInvalidResponseValue
from .ws import WS

logging.basicConfig(level=logging.DEBUG)
_LOGGER = logging.getLogger(__name__)


class PyPetWALK:
    """Class for communicate with the petWALK.control module."""

    def __init__(
        self, host: str, api_port: int = API_PORT, ws_port: int = WS_PORT
    ) -> None:
        """Initialize pyPetWALK Class."""
        self.websocket_client = WS(host, ws_port)
        self.api_client = API(host, api_port)

    async def __aenter__(self) -> "PyPetWALK":
        """Start pyPetWALK class from context manager."""
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc: BaseException | None,
        traceback: TracebackType | None,
    ) -> None:
        """Stop pyPetWALK class from context manager."""
        await self.websocket_client.close()
        await self.api_client.close()

    async def set_brightness_sensor(self, state: bool) -> bool:
        """Set new value for brightness sensor."""
        try:
            return await self.__api_set_state(API_STATE_BRIGHTNESS_SENSOR, state)
        finally:
            await self.api_client.close()

    async def get_brightness_sensor(self) -> bool:
        """Get current value for brightness sensor."""
        try:
            return await self.__api_get_state(API_STATE_BRIGHTNESS_SENSOR)
        finally:
            await self.api_client.close()

    async def set_motion_in(self, state: bool) -> bool:
        """Set new value for 'motion in' mode."""
        return await self.__api_set_state(API_STATE_MOTION_IN, state)

    async def get_motion_in(self) -> bool:
        """Get value for the 'motion in' mode."""
        return await self.__api_get_state(API_STATE_MOTION_IN)

    async def set_motion_out(self, state: bool) -> bool:
        """Set new value for 'motion out' mode."""
        return await self.__api_set_state(API_STATE_MOTION_OUT, state)

    async def get_motion_out(self) -> bool:
        """Get value for the 'motion out' mode."""
        return await self.__api_get_state(API_STATE_MOTION_OUT)

    async def set_rfid(self, state: bool) -> bool:
        """Set new value for rfid mode."""
        return await self.__api_set_state(API_STATE_RFID, state)

    async def get_rfid(self) -> bool:
        """Get value for the rfid mode."""
        return await self.__api_get_state(API_STATE_RFID)

    async def set_time(self, state: bool) -> bool:
        """Set new value for time mode."""
        return await self.__api_set_state(API_STATE_TIME, state)

    async def get_time(self) -> bool:
        """Get value for the time mode."""
        return await self.__api_get_state(API_STATE_TIME)

    async def set_door_state(self, state: bool) -> bool:
        """Open or closes petWALK door."""
        return await self.__api_set_state(API_STATE_DOOR, state)

    async def get_door_state(self) -> bool:
        """Get the current door state."""
        return await self.__api_get_state(API_STATE_DOOR)

    async def set_system_state(self, state: bool) -> bool:
        """Turn petWALK on or off."""
        return await self.__api_set_state(API_STATE_SYSTEM, state)

    async def get_system_state(self) -> bool:
        """Get current petWALK system state."""
        return await self.__api_get_state(API_STATE_SYSTEM)

    async def __api_get_state(self, param: str) -> bool:
        """Call API method to get the request mode/state."""
        method = f"get_{API_METHOD_MAPPING[param].lower()}s"
        _LOGGER.debug("Calling API method %s for %s", method, param)
        resp = await getattr(self.api_client, method)()
        if param not in resp:
            error = f"Invalid Response {param} not found in {{resp}}"
            _LOGGER.debug(error)
            raise PyPetWALKInvalidResponse(error)

        if resp[param] is True or resp[param] is False:
            result = resp[param]
        elif resp[param] in API_STATE_MAPPING:
            result = API_STATE_MAPPING[resp[param]]
        else:
            error = f"Unknown response value {resp[param]} for {param}"
            _LOGGER.error(error)
            raise PyPetWALKInvalidResponseValue(error)

        return result  # type: ignore[no-any-return]

    async def __api_set_state(self, param: str, value: bool) -> bool:
        """Call API method to set new value for requested mode/state."""
        method = f"set_{API_METHOD_MAPPING[param].lower()}"
        _LOGGER.debug(
            "Calling API method %s for %s with value %r", method, param, value
        )
        await getattr(self.api_client, method)(param, value)
        return True

    async def get_device_info(self) -> dict:
        """Get current device information."""
        return await self.websocket_client.device_info()



"""Module for pypetwalk Exceptions."""
from __future__ import annotations

from typing import Any


class BasePyPetWALKException(Exception):
    """pypetwalk Base exception."""

    def __init__(self, name: str, *args: Any) -> None:
        """Init the BasePyPetWALKException."""
        super().__init__(*args)
        self.name = name


class PyPetWALKClientConnectionError(BasePyPetWALKException):
    """pypetwalk PyPetWALKClientConnectionError exception."""

    def __init__(self, *args: Any) -> None:
        """Init the PyPetWALKClientConnectionError."""
        super().__init__("PyPetWALKClientConnectionError", *args)


class PyPetWALKInvalidResponse(BasePyPetWALKException):
    """pypetwalk PyPetWALKInvalidResponse exception."""

    def __init__(self, *args: Any) -> None:
        """Init the PyPetWALKInvalidResponse."""
        super().__init__("PyPetWALKInvalidResponse", *args)


class PyPetWALKInvalidResponseValue(BasePyPetWALKException):
    """pypetwalk PyPetWALKInvalidResponseValue exception."""

    def __init__(self, *args: Any) -> None:
        """Init the PyPetWALKInvalidResponseValue."""
        super().__init__("PyPetWALKInvalidResponseValue", *args)


class PyPetWALKInvalidResponseStatus(BasePyPetWALKException):
    """pypetwalk PyPetWALKInvalidResponseStatus exception."""

    def __init__(self, *args: Any) -> None:
        """Init the PyPetWALKInvalidResponseStatus."""
        super().__init__("PyPetWALKInvalidResponseStatus", *args)



"""pypetwalk is a Python library to communicate with the petWALK.control module."""
from __future__ import annotations

import json
import logging
from types import TracebackType

from aiohttp import ClientSession, WSMsgType
from aiohttp.client_exceptions import ClientConnectorError

from pypetwalk.const import (
    WS_COMMAND_DEVICE_INFO,
    WS_COMMAND_FACTORY_RESET,
    WS_COMMAND_INIT_DRIVE_START,
    WS_COMMAND_RFID_DELETE,
    WS_COMMAND_RFID_DELETE_ALL,
    WS_COMMAND_RFID_DELETE_PET,
    WS_COMMAND_RFID_START_LEARN,
    WS_COMMAND_RFID_STOP_LEARN,
    WS_COMMAND_RFID_TAG_EXISTS,
    WS_COMMAND_RFID_TAG_LIST,
    WS_COMMAND_TIME_SET,
    WS_COMMAND_WIFI_NETWORK_LIST,
    WS_COMMAND_WIFI_NETWORK_SET,
    WS_COMMAND_WIFI_SCAN,
    WS_COMMAND_ZIGBEE_JOIN_ALLOWED,
    WS_COMMAND_ZIGBEE_JOIN_CONFIRM,
    WS_COMMAND_ZIGBEE_LIST_DEVICES,
    WS_COMMAND_ZIGBEE_NAME_DEVICE,
    WS_COMMAND_ZIGBEE_REMOVE_DEVICE,
    WS_COMMAND_ZIGBEE_UPDATE,
    ZIGBEE_DEFAULT_JOIN_TYPE,
)
from pypetwalk.exceptions import PyPetWALKClientConnectionError

from .request import Request

_LOGGER = logging.getLogger(__name__)


class WS:
    """Class for Websocket communication."""

    def __init__(self, host: str, port: int) -> None:
        """Initialize Websocket Class."""
        self.server_host = host
        self.server_port = port
        self.session = ClientSession()

    async def __aenter__(self) -> "WS":
        """Start Websocket class from context manager."""
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc: BaseException | None,
        traceback: TracebackType | None,
    ) -> None:
        """Stop Websocket class from context manager."""
        await self.close()

    async def close(self) -> None:
        """Wait until all sessions are closed."""
        if self.session:
            await self.session.close()

    async def rfid_start_learn(self, slot: int) -> dict:
        """Start RFID learning process."""
        return await self.send_command(WS_COMMAND_RFID_START_LEARN, [slot])

    # @todo - Undocumented! - Need to find out correct parameters
    async def rfid_stop_learn(self) -> dict:
        """Stop RFID learning process."""
        return await self.send_command(WS_COMMAND_RFID_STOP_LEARN, [])

    async def rfid_delete(self, idx: int) -> dict:
        """Delete given RFID entry."""
        return await self.send_command(WS_COMMAND_RFID_DELETE, [idx])

    async def rfid_delete_all(self) -> dict:
        """Delete all RFID entries."""
        return await self.send_command(WS_COMMAND_RFID_DELETE_ALL, [])

    async def rfid_delete_pet(self, pet_id: str) -> dict:
        """Delete RFID based on petId."""
        return await self.send_command(WS_COMMAND_RFID_DELETE_PET, [pet_id])

    async def rfid_tag_list(self) -> dict:
        """Get list of current RFIDs."""
        return await self.send_command(WS_COMMAND_RFID_TAG_LIST, [])

    # @todo - Undocumented! - Need to find out correct parameters
    async def rfid_tag_exists(self) -> dict:
        """Check if RFID exists."""
        return await self.send_command(WS_COMMAND_RFID_TAG_EXISTS, [])

    async def zig_bee_list_devices(self) -> dict:
        """List ZigBee Devices."""
        return await self.send_command(WS_COMMAND_ZIGBEE_LIST_DEVICES, [])

    async def zig_bee_remove_device(self, component_id: str) -> dict:
        """Remove ZigBee Device."""
        return await self.send_command(WS_COMMAND_ZIGBEE_REMOVE_DEVICE, [component_id])

    async def zig_bee_start_join(self, z_type: str) -> dict:
        """Start ZigBee joining session."""
        if not z_type:
            z_type = ZIGBEE_DEFAULT_JOIN_TYPE
        return await self.send_command(
            WS_COMMAND_ZIGBEE_JOIN_ALLOWED, ["start", z_type]
        )

    async def zig_bee_get_join_status(self) -> dict:
        """Check current ZigBee joining status."""
        return await self.send_command(WS_COMMAND_ZIGBEE_JOIN_ALLOWED, [])

    async def zig_bee_join_confirm(self, component_id: str) -> dict:
        """Confirm ZigBee join for given component."""
        return await self.send_command(WS_COMMAND_ZIGBEE_JOIN_CONFIRM, [component_id])

    # @todo - Undocumented! - Need to find out correct parameters
    async def zig_bee_update(self) -> dict:
        """Update ZigBee status."""
        return await self.send_command(WS_COMMAND_ZIGBEE_UPDATE, [])

    async def zig_bee_name_device(self, component_id: str, name: str) -> dict:
        """Set name for ZigBee Device."""
        return await self.send_command(
            WS_COMMAND_ZIGBEE_NAME_DEVICE, [component_id, name]
        )

    async def device_info(self) -> dict:
        """Get current device information."""
        return await self.send_command(WS_COMMAND_DEVICE_INFO, [])

    async def wifi_network_list(self) -> dict:
        """Get Wifi network list."""
        return await self.send_command(WS_COMMAND_WIFI_NETWORK_LIST, [])

    async def wifi_network_set(
        self,
        selected_wifi: str,
        wifi_password: str,
    ) -> dict:
        """Set Wifi network."""
        return await self.send_command(
            WS_COMMAND_WIFI_NETWORK_SET, [selected_wifi, wifi_password, False]
        )

    # @todo - Undocummented! - Need to find out correct parameters
    async def wifi_scan(self) -> dict:
        """Start Wifi scan."""
        return await self.send_command(WS_COMMAND_WIFI_SCAN, [])

    # @todo - Time format is undocumented!
    async def time_set(self, time: str) -> dict:
        """Set given time."""
        return await self.send_command(WS_COMMAND_TIME_SET, [time])

    async def factory_reset(self) -> dict:
        """Reset the device to factory settings."""
        _LOGGER.warning("Factory reset was triggered!")
        return await self.send_command(WS_COMMAND_FACTORY_RESET, [])

    async def init_drive_start(self) -> dict:
        """Init drive start for the device."""
        _LOGGER.warning("Init Drive Start was triggered!")
        return await self.send_command(WS_COMMAND_INIT_DRIVE_START, [])

    async def send_command(self, command: str, params: list) -> dict:
        """Send command to local Websocket."""
        request = Request().build_request(command, params)

        url = f"ws://{self.server_host}:{self.server_port}"
        try:
            async with self.__get_session().ws_connect(url) as websocket_connection:
                await websocket_connection.send_str(request.get_json())

                async for msg in websocket_connection:
                    if msg.type == WSMsgType.ERROR:
                        _LOGGER.error("Unable to connect to WS %s", url)
                        result = {}
                    else:
                        if msg.type == WSMsgType.TEXT:
                            result = json.loads(msg.data)
                    return result
        except ClientConnectorError as ex:
            _LOGGER.debug("%s", ex)
            await self.close()
            raise PyPetWALKClientConnectionError(ex) from ex

        return {}

    def __get_session(self) -> ClientSession:
        """Return current session, recreating if it was closed."""
        if self.session.closed:
            self.session = ClientSession()

        return self.session
